#!/usr/bin/wish

#
#	Mouse-stroke character recognizer,
#	major pieces stolen from other sources.
#	Mike Hall, mghall@enteract.com, 2000-12-23
#

#
#	mouse stroke collector, from page 484 of
#	Practical Programming in Tcl/Tk, Third Edition, 2000, Brent Welch
#

proc StrokeInit {w} {
	bind $w <Button-1>	  {StrokeBegin	%W %x %y}
	bind $w <B1-Motion>	  {Stroke	%W %x %y}
	bind $w <ButtonRelease-1> {StrokeEnd	%W %x %y}
	return
}

proc StrokeBegin {w x y} {
	global stroke
	catch {unset stroke}
	set stroke(N) 0
	set stroke(0) [list $x $y]
	msg "1 point ..."
	return
}

proc Stroke {w x y} {
	global stroke

	# get last point
	set n $stroke(N)
	foreach {ox oy} $stroke($n) {break}

	# filter? abs(dx) + abs(dy) > threshold

	# install latest point
	incr n
	set stroke(N) $n
	set stroke($n) [list $x $y]
	msg "$n points ..."

	# draw latest segment
	$w create line $ox $oy $x $y -width 2 -tag segments
	return
}

proc StrokeEnd {w x y} {
	global stroke

	set n $stroke(N)
	msg "$n points"

	if {$n > 4} {
		# arrow at end
		foreach {ox oy} $stroke([expr {$n - 4}]) {break}
		$w create line $ox $oy $x $y -arrow last -tag segments

		# if enough points, process it
		recog
	}

	# dot at begin
	foreach {ox oy} $stroke(0) {break}
	set  x [expr {$ox + 4}]
	set  y [expr {$oy + 4}]
	set ox [expr {$ox - 4}]
	set oy [expr {$oy - 4}]
	$w create oval $ox $oy $x $y -fill white -tag segments
	return
}

#
#	Get min/max x/y values of the stroke
#
proc get_min_max { } {
	global stroke
	global xl xh yl yh x1 x2 y1 y2

	# initialize from first point
	foreach {xl yl} $stroke(0) {break}
	set xh $xl
	set yh $yl

	# adjust from remaining data
	set n $stroke(N)
	for {set i 1} {$i <= $n} {incr i} {
		foreach {x y} $stroke($i) {break}
		if {$x < $xl} {
			set xl $x
		} elseif {$x > $xh} {
			set xh $x
		}
		if {$y < $yl} {
			set yl $y
		} elseif {$y > $yh} {
			set yh $y
		}
	}

	# divide the box in thirds each way (9 sub boxes)
	set x1 [expr {$xl + ($xh - $xl)/3.}]
	set y1 [expr {$yl + ($yh - $yl)/3.}]
	set x2 [expr {$xl + 2.*($xh - $xl)/3.}]
	set y2 [expr {$yl + 2.*($yh - $yl)/3.}]

	# check aspect (for vertical and horizontal strokes)
	set dx [expr {abs($xh - $xl)}]
	set dy [expr {abs($yh - $yl)}]
	set thresh 6.0
	if {$dy > [expr {$thresh * $dx}]} {
		# vertical
		set x1 $xl
		set x2 $xh
	} elseif {$dx > [expr {$thresh * $dy}]} {
		# horizontal
		set y1 $yl
		set y2 $yh
	}

	return
}

#
#	Display the box outline and the interior dividers
#
proc show_boxes { } {
	global stroke
	global xl xh yl yh x1 x2 y1 y2

	# enclosing box
	.c create rectangle [expr {$xl-1}] [expr {$yl-1}]  \
		[expr {$xh+1}] [expr {$yh+1}] -tags boxes -outline blue

	# interior crossing lines
	.c create line $xl $y1 $xh $y1 -tags boxes -fill red
	.c create line $xl $y2 $xh $y2 -tags boxes -fill red
	.c create line $x1 $yl $x1 $yh -tags boxes -fill red
	.c create line $x2 $yl $x2 $yh -tags boxes -fill red

	# label
	.c create text $xl $yl -text $stroke(N) -anchor sw -fill Yellow \
		-tags boxes
	return
}

#
#	Convert from x/y coordinates to cell-values in the box
#	
#	Normal		XOR crossings
#
#	0 1 2		 0  1  3
#	3 4 5		 4  5  7
#	6 7 8		12 13 15
#
proc cell_value {x y} {
	global x1 x2 y1 y2

	# get cell value for x coordinate
	if {$x < $x1} {
		set xv 0
	} elseif {$x <= $x2} {
		set xv 1
	} else {
		set xv 3
	}

	# get cell value for y coordinate
	if {$y < $y1} {
		set yv 0
	} elseif {$y <= $y2} {
		set yv 4
	} else {
		set yv 12
	}

	# overall cell value
	return [expr {$xv + $yv}]
}

#
#	Reset crossing counts
#
proc init_crossing { } {
	global crosses
	foreach x {1 2 4 8 N} {
		set crosses($x) 0
	}
	return
}

#
#	Track line-crossing counts from old-cell to new-cell
#
proc crossing {old new} {
	global crosses
	incr crosses(N)
	set cn [expr {$new ^ $old}]
	foreach bit {1 2 4 8} {
		if {$cn & $bit} {
			incr crosses($bit)
		}
	}
	return
}

#
#	Convert strokes to list of cells
#
proc box_cells { } {
	global stroke 
	global cells

	# setup
	catch {unset cells}
	init_crossing

	# start first point
	foreach {x y} $stroke(0) {break}
	set ocv [cell_value $x $y]
	set cells $ocv

	# convert rest of points
	set n $stroke(N)
	for {set i 1} {$i <= $n} {incr i} {
		foreach {x y} $stroke($i) {break}
		set cv [cell_value $x $y]
		if {$cv != $ocv} {
			# new cell, accumulate
			lappend cells $cv
			# track line crossings
			crossing $ocv $cv
			# new current cell value
			set ocv $cv
		}
	}
	return
}

#
#	Construct the resulting set of features
#		- first cell, last cell
#		- last crossing, last crossing
#		- four crossing counts (2 x-axis, 2 y-axis)
#
proc ftr { } {
	global cells
	global crosses

	set c0 [lindex $cells 0]
	set c1 [lindex $cells 1]
	set cn [lindex $cells end]
	set cp [lindex $cells [expr {[llength $cells]-2}]]

	set d1 [expr {$c0 ^ $c1}]
	set dn [expr {$cn ^ $cp}]

	return "$c0 $cn $d1 $dn $crosses(1) $crosses(2) $crosses(4) $crosses(8)"
}

proc show_ftr {f} {
	global xl xh yl yh
	global cells

	.c create text $xl $yh -text $f -anchor nw -fill Green -tags boxes
	set x [match $f]
	if {$x == ""} {
		set x unk
	}
	set x "[llength $cells] $x"
	msg $x
	.c create text $xh $yl -text $x -anchor se -fill Yellow -tags boxes

	return
}

#
#	Recognize the stroke
#	Leeden recognizer, from page 204 of
#	Principles of Interactive Computer Graphics 2nd ed, 1979, Newman+Sproull
#
proc recog { } {
	get_min_max
	show_boxes
	box_cells
	set f [ftr]
	show_ftr $f
	return
}

#
#	Clear the canvas, prepare for another run
#
proc clear { } {
	global stroke
	.c delete segments
	.c delete boxes
	unset stroke
	return
}

#
#	Update the log message
#
proc msg {x} {
	global msg
	set msg $x
	return
}

#
#	Lookup this stroke's features
#
proc match {f} {
	global ftrs

	if {[info exists ftrs($f)]} {
		return $ftrs($f)
	} else {
		return ""
	}
}

#
#	Set a new definition
#
proc set_def { } {
	global ftrs char

	set f [ftr]	
	set x [match $f]
	if {$x != ""} {
		msg "Conflict: $ftrs($f) and $char"
	} else { 
		set ftrs($f) $char
		msg "set $char"
	}
	return
}

#
#	Unset an existing definition
#
proc unset_def { } {
	global ftrs
	
	set f [ftr]
	if {[info exists ftrs($f)]} {
		set x $ftrs($f)
		msg "Erasing $x $f"
		unset ftrs($f)
	}
	return
}

#
#	Unset all definitions for a character
#
proc undef_char { } {
	global ftrs char
	foreach f [array names ftrs] {
		if {$ftrs($f) == $char} {
			unset ftrs($f)
		}
	}
	return
}

#
#	Read the features
#
proc read_defs { } {
	global ftrs
	global ftrfile

	if {[catch {open $ftrfile r} fid] == 0} {
		while {[gets $fid line] > 0} {
			set ftrs([lrange $line 1 end]) [lindex $line 0]
		}
		close $fid
	}
	msg "read [array size ftrs] entries"
	return
}

#
#	Save the defined features
#
proc save_defs { } {
	global ftrs
	global ftrfile

	if {[catch {open $ftrfile w} fid] == 0} {
		foreach {k v} [array get ftrs] {
			# escape some special characters
			if {[string match {[\{\[\" ]} $v]} {
				puts $fid "\\$v $k"
			} else {
				puts $fid "$v $k"
			}
		}
		close $fid
	}
	msg "saved [array size ftrs] entries"
	return
}



#
#	Create the toplevel interface
#
proc main {} {
	global msg char

	wm title . "Character Recognizer"

	# drawing surface
	canvas .c

	# row of controls
	frame .b
	button .b.d -text Done  -command exit
	button .b.r -text Read  -command read_defs
	button .b.s -text Save  -command save_defs
	button .b.c -text Erase -command clear
	button .b.u -text Unset -command unset_def
	button .b.f -text UnDef -command undef_char
	button .b.t -text Def   -command set_def
	entry  .b.e -width 3 -textvariable char
	bind .b.e <Return> {set_def}
	
	label .b.msg -textvariable msg -width 20
	pack .b.d .b.r .b.s .b.c .b.u .b.f .b.t .b.e -side left -fill x
	pack .b.msg -side right -fill x -expand 1

	# buttons along the bottom, canvas fills remainder
	pack .b -side bottom -fill x
	pack .c -side top -fill both -expand 1

	# setup stroke collector
	StrokeInit .c
	return
}


set ftrfile ftrs.def
main
read_defs
